from typing import Optional

import torch


class CollatorWithPadding():

    def __init__(
            self, 
            pad_to_n_support_samples: Optional[int]
        ) -> None:
        
        self.pad_to_n_support_samples = pad_to_n_support_samples


    def __call__(self, batch: list[dict[str, torch.Tensor]]) -> dict[str, torch.Tensor]:

        if self.pad_to_n_support_samples is not None:
            assert all(dataset['x_support'].shape[0] <= self.pad_to_n_support_samples for dataset in batch)
            self.n_support_samples = self.pad_to_n_support_samples
        else:
            self.n_support_samples = max(dataset['x_support'].shape[0] for dataset in batch)

        max_query_samples = max(dataset['x_query'].shape[0] for dataset in batch)

        n_support_features = batch[0]['x_support'].shape[1]
        n_query_features = batch[0]['x_query'].shape[1]
        y_dtype = batch[0]['y_support'].dtype

        batch_size = len(batch)

        tensor_dict = {
            'x_support': torch.zeros((batch_size, self.n_support_samples, n_support_features), dtype=torch.float32),
            'y_support': torch.zeros((batch_size, self.n_support_samples), dtype=y_dtype),
            'x_query': torch.zeros((batch_size, max_query_samples, n_query_features), dtype=torch.float32),
            'y_query': torch.zeros((batch_size, max_query_samples), dtype=y_dtype)
        }

        for i, dataset in enumerate(batch):
            tensor_dict['x_support'][i, :dataset['x_support'].shape[0], :] = dataset['x_support']
            tensor_dict['y_support'][i, :dataset['y_support'].shape[0]] = dataset['y_support']
            tensor_dict['x_query'][i, :dataset['x_query'].shape[0], :] = dataset['x_query']
            tensor_dict['y_query'][i, :dataset['y_query'].shape[0]] = dataset['y_query']

        return tensor_dict




from __future__ import annotations

from pathlib import Path

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
import torch

from .core.dataset_split import make_stratified_dataset_split
from .core.trainer_finetune import TrainerFinetune
from .models.foundation.foundation_transformer import FoundationTransformer


# FIXME: Delete this class, it isn't needed
# TODO: test_epoch might be better if it uses the for loop logic with n_ensembles during finetuning to better estimate val score
# TODO: To mitigate val overfitting, can fit multiple random seeds at same time and pick same epoch for all of them, track average performance on epoch.
# TODO: Test shuffling the data and see if it makes TabPFNv2 worse, same with TabForestPFN
class TabPFNMixClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, n_classes, cfg, split_val, model_path: str = None, weights_path: str | Path = None, stopping_metric=None, use_best_epoch: bool = True):
        if weights_path is not None:
            weights_path = str(Path(weights_path))

        if model_path is not None:
            model = FoundationTransformer.from_pretrained(model_path)
            assert model.task == cfg.task, f"The pretrained model '{model_path}' is for task {model.task}, but the problem type is for task {cfg.task}..."
        else:
            model = FoundationTransformer(
                n_features=cfg.hyperparams['n_features'],
                n_classes=cfg.hyperparams['n_classes'],
                dim=cfg.hyperparams['dim'],
                n_layers=cfg.hyperparams['n_layers'],
                n_heads=cfg.hyperparams['n_heads'],
                attn_dropout=cfg.hyperparams['attn_dropout'],
                y_as_float_embedding=cfg.hyperparams['y_as_float_embedding'],
                task=cfg.task,
            )
        if weights_path is not None:
            model.load_state_dict(torch.load(weights_path, weights_only=True))  # nosec B614

        self.split_val = split_val
        self.trainer = TrainerFinetune(cfg, model, n_classes=n_classes, stopping_metric=stopping_metric, use_best_epoch=use_best_epoch)
        super().__init__()

    def fit(self, X: np.ndarray, y: np.ndarray, X_val: np.ndarray = None, y_val: np.ndarray = None, time_limit: float = None):
        # FIXME: Should X and y be preprocessed for inference efficiency? Yes.
        self.X_ = X  # FIXME: Optimize storage of X and y? Is this redundant? Is X and y saving done multiple times during pickle?
        self.y_ = y

        if X_val is not None and y_val is not None:
            X_train, X_valid, y_train, y_valid = X, X_val, y, y_val
        elif self.split_val:
            X_train, X_valid, y_train, y_valid = make_stratified_dataset_split(X, y)  # FIXME: Add random seed
        else:
            X_train, X_valid, y_train, y_valid = X, None, y, None
        self.trainer.train(x_train=X_train, y_train=y_train, x_val=X_valid, y_val=y_valid, time_limit=time_limit)

        return self

    # FIXME: Avoid preprocessing self.X_ and self.y_ each predict call
    def predict(self, X):
        logits = self.trainer.predict(self.X_, self.y_, X)
        return logits.argmax(axis=1)

    # FIXME: Avoid preprocessing self.X_ and self.y_ each predict_proba call
    def predict_proba(self, X):
        logits = self.trainer.predict(self.X_, self.y_, X)
        return np.exp(logits) / np.exp(logits).sum(axis=1)[:, None]


import copy
import logging

import pandas as pd

logger = logging.getLogger(__name__)


# TODO: currently is buggy
class FeaturePruner:
    def __init__(self, model_base, threshold_baseline=0.004, is_fit=False):
        self.model_base = model_base
        self.threshold_baseline = threshold_baseline
        self.is_fit = is_fit

        self.best_score = 0
        self.best_iteration = 0
        self.features_in_iter = []
        self.score_in_iter = []
        self.valid_feature_counts = []
        self.gain_dfs = []
        self.banned_features_in_iter = []
        self.thresholds = []
        self.threshold = self.threshold_baseline
        self.cur_iteration = 0
        self.tuned = False
        self.early_stopping_rounds = 2

    def evaluate(self):
        untuned_score = self.score_in_iter[0]
        logger.debug("untuned_score: %s" % untuned_score)
        logger.debug("best_score: %s" % self.best_score)
        logger.debug("best_iteration: %s" % self.best_iteration)

        data_dict = {
            "score": self.score_in_iter,
            "feature_count": self.valid_feature_counts,
            "threshold": self.thresholds,
        }
        logger.debug(self.gain_dfs[self.best_iteration])
        z = pd.DataFrame(data=data_dict)
        # logger.debug(z)
        z_sorted = z.sort_values(by=["score"], ascending=False)
        # logger.debug(z_sorted)

    # TODO: CV5 instead of holdout? Should be better
    # TODO: Add holdout here, it is overfitting with Logistic Regression
    def tune(self, X, y, X_val, y_val, X_holdout, y_holdout, total_runs=999):
        objective_goal_is_negative = False  # Fixed to false if using sklearn scorers # self.model_base.problem_type == REGRESSION  # TODO: if objective function goal = lower (logloss, MAE, etc.)
        logger.log(15, "Feature-pruning " + str(self.model_base.name) + " for " + str(total_runs) + " runs...")

        if len(self.features_in_iter) == 0:
            valid_features = X.columns.values
        else:
            valid_features = self.features_in_iter[-1]

        iter_since_best = 0
        iter_start = self.cur_iteration
        for iteration in range(self.cur_iteration, total_runs):
            self.cur_iteration = iteration
            logger.debug("iteration: %s " % iteration)
            X_subset = X[valid_features].copy()
            X_val_subset = X_val[valid_features].copy()
            self.thresholds.append(self.threshold)
            self.valid_feature_counts.append(len(valid_features))
            self.features_in_iter.append(valid_features)

            if self.is_fit and (iteration == iter_start):
                model_iter = self.model_base
            else:
                model_iter = copy.deepcopy(self.model_base)
                model_iter.fit(X=X_subset, y=y, X_val=X_val_subset, y_val=y_val)

            banned_features = []

            feature_importance = None
            if hasattr(model_iter.model, "feature_importances_"):
                feature_importance = model_iter.model.feature_importances_
            elif hasattr(model_iter.model, "feature_importance"):
                feature_importance = model_iter.model.feature_importance()

            if feature_importance is not None:
                a = pd.Series(data=feature_importance, index=X_subset.columns)
                unused_features = a[a == 0]

                logger.log(15, "Unused features after pruning: " + str(list(unused_features.index)))
                features_to_use = [feature for feature in valid_features if feature not in unused_features.index]
                banned_features += list(unused_features.index)
            else:
                features_to_use = list(valid_features)

            cur_score_val = model_iter.score(X=X_val_subset, y=y_val)
            cur_score = model_iter.score(X=X_holdout[valid_features], y=y_holdout)

            logger.log(15, "Iter " + str(iteration) + "  Score: " + str(cur_score))
            logger.log(15, "Iter " + str(iteration) + "  Score Val: " + str(cur_score_val))
            if objective_goal_is_negative:
                cur_score = -cur_score

            if self.best_score == 0 or cur_score > self.best_score:
                logger.log(15, "New best score found!")
                logger.log(15, str(cur_score) + " > " + str(self.best_score))
                self.best_score = cur_score
                self.best_iteration = iteration
                iter_since_best = 0
            else:
                iter_since_best += 1

            self.score_in_iter.append(cur_score)

            if iter_since_best >= self.early_stopping_rounds:
                logger.log(15, "Early stopping on iter %s" % iteration)
                logger.log(15, "Best Iteration: %s" % self.best_iteration)
                self.tuned = True
                break

            gain_df = model_iter.compute_feature_importance(X=X_val_subset, y=y_val, features=features_to_use)
            if not objective_goal_is_negative:
                gain_df = -gain_df

            self.gain_dfs.append(gain_df)

            # TODO: Save gain_df, banned_features

            self.threshold = self.adjust_threshold(gain_df, self.threshold)

            banned_df = gain_df.loc[gain_df >= self.threshold].index
            banned_features += list(banned_df)

            logger.log(15, "Banned features: " + str(banned_features))
            self.banned_features_in_iter.append(banned_features)
            valid_features = [feature for feature in valid_features if feature not in banned_features]

            if len(valid_features) == 0:
                logger.log(15, "No more features to remove, Feature pruning complete")
                logger.log(15, "score_in_iter: " + str(self.score_in_iter))
                self.tuned = True
                break
        self.tuned = True

    @staticmethod
    def adjust_threshold(gain_df, threshold):
        if gain_df.shape[0] == 0:
            raise BaseException("ran out of features to prune!")
        banned_df = gain_df.loc[gain_df >= threshold].index
        banned_features = list(banned_df)

        if len(banned_features) == 0:
            if threshold < -100000000:  # FIXME: Hacked for regression
                raise BaseException("threshold already max!")
            elif threshold > 0.000001:
                threshold_new = threshold / 2
            elif threshold > 0:
                threshold_new = 0
            elif threshold == 0:
                threshold_new = -0.00000001
            elif (threshold < 0) and (threshold > -0.0001):
                threshold_new = threshold * 2
            else:
                threshold_new = gain_df.max()

            logger.log(10, "Adjusting threshold to %s" % threshold_new)
            return FeaturePruner.adjust_threshold(gain_df=gain_df, threshold=threshold_new)
        else:
            return threshold



from __future__ import annotations

from dataclasses import dataclass

import torch

from ..core.enums import Task


@dataclass
class ConfigRun:
    task: Task
    hyperparams: dict
    seed: int = 0
    device: torch.device = None



from autogluon.core.constants import BINARY, MULTICLASS, QUANTILE, REGRESSION


# TODO this method is generalizable and potentially should be moved out into framework
def get_param_baseline(problem_type, num_classes=None):
    if problem_type == BINARY:
        return get_param_binary_baseline()
    elif problem_type == MULTICLASS:
        return get_param_multiclass_baseline()
    elif problem_type == REGRESSION:
        return get_param_regression_baseline()
    elif problem_type == QUANTILE:
        return get_param_quantile_baseline()
    else:
        return get_param_binary_baseline()


def get_param_multiclass_baseline():
    # TODO: explore/add other hyperparameters like weight decay, use of batch-norm, activation-function choice, etc.
    params = {
        # See docs: https://docs.fast.ai/tabular.learner.html
        "layers": None,  # layers configuration; None - use model's heuristics
        "emb_drop": 0.1,  # embedding layers dropout
        "ps": 0.1,  # linear layers dropout
        "bs": "auto",  # batch size
        # maximum learning rate for one cycle policy https://docs.fast.ai/train.html#fit_one_cycle
        # One-cycle policy paper: https://arxiv.org/abs/1803.09820
        "lr": 1e-2,
        "epochs": "auto",  # maximum number of epochs
        # Early stopping settings. See more details here: https://docs.fast.ai/callbacks.tracker.html#EarlyStoppingCallback
        "early.stopping.min_delta": 0.0001,
        "early.stopping.patience": 20,
        # If > 0, then use LabelSmoothingCrossEntropy loss function for binary/multi-class classification;
        # otherwise use default loss function for this type of problem
        "smoothing": 0.0,
    }
    return params


def get_param_binary_baseline():
    return get_param_multiclass_baseline()


def get_param_regression_baseline():
    return get_param_multiclass_baseline()


def get_param_quantile_baseline():
    params = get_param_regression_baseline()

    # residual threshold parameter in HuberPinballLoss
    params.update({"alpha": 0.01})
    return params
